// ***************************************************************************
// ** Datengenerierung                                                      **
// ***************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <vector>
#include <set>


#ifdef CYGWIN
	#define MAXINT 2147483647
#else
	#define MAXINT 2147483647 //32767
#endif

#define sqr(a) ((a)*(a))

using namespace std;

long Statistics_Count;
long num_Scale;
double* Statistics_SumX;
double* Statistics_SumXsquared;
double* Statistics_SumProduct;
vector<double> *poissonDistribution;
double discount, variance;

void InitStatistics(long Dimensions)
//   ==============
// initialisiert Z�hlvariablen der Statistik
{
	Statistics_SumX = new double[Dimensions];
	Statistics_SumXsquared = new double[Dimensions];
	Statistics_SumProduct = new double[Dimensions*Dimensions];

	Statistics_Count = 0;
	for (long d=0; d<Dimensions; d++) {
		Statistics_SumX[d]=0.0;
		Statistics_SumXsquared[d]=0.0;
		for (long dd=0; dd<Dimensions; dd++) Statistics_SumProduct[d*Dimensions+dd] = 0.0;
	}
}


void EnterStatistics(long Dimensions,double* x)
//   ===============
// registiriert den Vektor "x" f�r die Statistik
{
	Statistics_Count++;
	for (long d=0; d<Dimensions; d++) {
		Statistics_SumX[d] += x[d];
		Statistics_SumXsquared[d] += x[d]*x[d];
		for (long dd=0; dd<Dimensions; dd++) Statistics_SumProduct[d*Dimensions+dd] += x[d]*x[dd];
	}
}


void OutputStatistics(long Dimensions)
//   ================
// gibt die Statistik aus
{
	long d;
	for (d=0; d<Dimensions; d++) {
		double E = Statistics_SumX[d] / Statistics_Count;
		double V = Statistics_SumXsquared[d]/Statistics_Count - E*E;
		double s = sqrt(V);
		printf("E[X%d]=%5.2f  Var[X%d]=%5.2f  s[X%d]=%5.2f\n",d+1,E,d+1,V,d+1,s);
	}
	printf("\nKorrelationsmatrix:\n");
	for (d=0; d<Dimensions; d++) {
		for (long dd=0; dd<Dimensions; dd++) {
			double Kov = (Statistics_SumProduct[d*Dimensions+dd]/Statistics_Count) -
			             (Statistics_SumX[d]/Statistics_Count) * (Statistics_SumX[dd]/Statistics_Count);
			double Cor = Kov /
			             sqrt(Statistics_SumXsquared[d]/Statistics_Count - sqr(Statistics_SumX[d] / Statistics_Count)) /
			             sqrt(Statistics_SumXsquared[dd]/Statistics_Count - sqr(Statistics_SumX[dd] / Statistics_Count));
			printf("  %5.2f",Cor);
		}
		printf("\n");
	}
	printf("\n");
}


double RandomEqual(double min,double max)
//     ===========
// liefert eine im Intervall [min,max[ gleichverteilte Zufallszahl
{
	long k = rand();
	//printf("%d ", k);

	double x = (double)k/MAXINT;

	return x*(max-min)+min;
}


double RandomPeak(double min,double max,long dim)
//     ==========
// liefert eine Zufallsvariable im Intervall [min,max[
// als Summe von "dim" gleichverteilten Zufallszahlen
{
	double sum = 0.0;
	for (long d=0; d<dim; d++) sum += RandomEqual(0,1);
	sum /= dim;
	return sum*(max-min)+min;
}


double RandomNormal(double med,double var)
//     ============
// liefert eine normalverteilte Zufallsvariable mit Erwartungswert med
// im Intervall ]med-var,med+var[
{
	return RandomPeak(med-var,med+var,12);
}


double **GenerateDataEqually(long Count,long Dimensions)
//   ===================
// generiert in der Datei "f" "Count" gleichverteilte Datens�tze
{
	double **returnValue;

	returnValue = (double **) malloc(sizeof(double *)*Count);
	InitStatistics(Dimensions);
	for (long i=0; i<Count; i++)
	{
		//double x[Dimensions];
		double *x;
		x = (double *) malloc(sizeof(double)*Dimensions);
		for (long d=0; d<Dimensions; d++)
		{
			x[d] = RandomEqual(0,1);
			//fprintf(f,"%8.6f ",x[d]);
		}
		EnterStatistics(Dimensions,x);
		//fprintf(f,"\n");

		returnValue[i] = x;
	}
	OutputStatistics(Dimensions);

	return returnValue;
}


double **GenerateDataCorrelated(long Count,long Dimensions)
//   ======================
// generiert in der Datei "f" "Count" korrelierte Datens�tze
{
	double **returnValue;
	long d;
	long i;

	returnValue = (double **) malloc(sizeof(double *)*Count);

	InitStatistics(Dimensions);
	//double x[Dimensions];

	for (i=0; i<Count; i++) {
		double *x;
		x = (double *) malloc(sizeof(double)*Dimensions);
again:
		double v = RandomPeak(0,1,Dimensions);
		for (d=0; d<Dimensions; d++) x[d] = v;
		double l = v<=0.5 ? v:1.0-v;
		for (d=0; d<Dimensions; d++) {
			double h = RandomNormal(0,l);
			x[d] += h;
			x[(d+1)%Dimensions] -= h;
		}
		for (d=0; d<Dimensions; d++) if (x[d]<0 || x[d]>=1) goto again;
		//for (d=0; d<Dimensions; d++) fprintf(f,"%8.6f ",x[d]);
		EnterStatistics(Dimensions,x);
		//fprintf(f,"\n");

		returnValue[i] = x;
	}
	OutputStatistics(Dimensions);

	return returnValue;
}


double **GenerateDataAnticorrelated(long Count,long Dimensions)
//   ==========================
// generiert in der Datei "f" "Count" antikorrelierte Datens�tze
{
	double **returnValue;

	returnValue = (double **) malloc(sizeof(double *)*Count);

	InitStatistics(Dimensions);
	//double x[Dimensions];
	double *x;
	long d;
	long i;
	bool flag;

	for (i=0; i<Count; i++) {

		x = (double *) malloc(sizeof(double)*Dimensions);

		do
		{

		flag = false;
		double v = RandomNormal(0.5,0.25);
		for (d=0; d<Dimensions; d++) x[d] = v;
		double l = (v<=0.5) ? v:(1.0-v);
		for (d=0; d<Dimensions; d++) {
			double h = RandomEqual(-l,l);
			x[d] += h;
			x[(d+1)%Dimensions] -= h;
		}
		for (d=0; d<Dimensions; d++){
			if (x[d]<0 || x[d]>=1){
				flag = true;
				break;
			}
		}

		}while(flag);
		//for (d=0; d<Dimensions; d++) fprintf(f,"%8.6f ",x[d]);
		EnterStatistics(Dimensions,x);
		//fprintf(f,"\n");

		returnValue[i] = x;
	}
	OutputStatistics(Dimensions);

	return returnValue;
}


double *generateZipProb(double theta, long noOfElt)
{
	double *returnValue;
	long i;

	double total;

	long rank;
	double A;

	returnValue = (double *) malloc(sizeof(double)*noOfElt);
	memset(returnValue, 0, sizeof(double)*noOfElt);

	total = 0;
	for (i = 0; i < noOfElt; i++)
	{
		rank = i+1;
		total += 1.0/pow(rank, theta);
	}

	A = 1.0/total;

	for (i = 0; i < noOfElt; i++)
	{
		rank = i+1;
		returnValue[i] = A/pow(rank, theta);
	}

	return returnValue;
}

long generateRandomClass(long noOfElt, double *prob)
{
	long i;

	double *accum;

	long returnValue;

	double min, max;
	double randNo;

	accum = (double *) malloc(sizeof(double)*noOfElt);
	memset(accum, 0, sizeof(double)*noOfElt);

	accum[0] = prob[0];
	for (i = 1; i < noOfElt; i++)
	{
		accum[i] = accum[i-1]+prob[i];
	}

	min = 0;
	max = accum[noOfElt-1];

	randNo = RandomEqual(min, max);

	returnValue = -1;
	for (i = 0; i < noOfElt; i++)
	{
		if (randNo <= accum[i])
		{
			returnValue = i;
			break;
		}
	}

	free(accum);

	return returnValue;
}

//catData = GenerateCatData(Count, noOfCatDim, noOfClass);
long **GenerateCatData(long noOfData, long noOfCatDim, long noOfClass, double theta)
{
	long **returnValue;
	long i, j;

	double *prob;

	long classNo;

	if (noOfCatDim > 0)
	{
		returnValue = (long **) malloc(sizeof(long *)*noOfData);

		for (i = 0; i < noOfData; i++)
		{
			returnValue[i] = (long *) malloc(sizeof(int)*noOfCatDim);
		}

		// generate prob.
		prob = generateZipProb(theta, noOfClass);

		for (i = 0; i < noOfData; i++)
		{
			for (j = 0; j < noOfCatDim; j++)
			{
				classNo = generateRandomClass(noOfClass, prob);
				returnValue[i][j] = classNo;
			}
		}

		free(prob);
	}
	else
	{
		returnValue = NULL;
	}

	return returnValue;
}

void OutputData(FILE *filePtr, long noOfData, double **numData, long noOfNumDim)
{
	long i, j;

	long aNo;

	long scale = 10000;

	for (i = 0; i < noOfData; i++)
	{

		for (j = 0; j < noOfNumDim; j++)
		{
			if (numData[i][j] == 1.0)
			{
				aNo = scale-1;
			}
			else
			{
				aNo = (long) (numData[i][j]*scale);
			}
			fprintf(filePtr, "%ld", aNo);
			if (j != noOfNumDim -1)
			{
				fprintf(filePtr, " ");
			}
		}

		fprintf(filePtr, "\n");
	}

}


void OutputData3(FILE *filePtr, long noOfData, double **numData, long noOfNumDim)
{
	long i, j;

	long aNo;

	long* scale = new long[noOfNumDim];

	for(i=0; i < noOfNumDim - 1; i++)
	{
		scale[i] = (long)RandomNormal(50.0,0.25);
	}

	scale[noOfNumDim - 1] = 10000;

	for (i = 0; i < noOfData; i++)
	{

		for (j = 0; j < noOfNumDim; j++)
		{
			if (numData[i][j] == 1.0)
			{
				aNo = scale[j]-1;
			}
			else
			{
				aNo = (long) (numData[i][j]*scale[j]);
			}
			fprintf(filePtr, "%ld", aNo);
			if (j != noOfNumDim -1)
			{
				fprintf(filePtr, " ");
			}
		}

		fprintf(filePtr, "\n");
	}
}


void OutputData3(FILE *filePtr, long noOfData, double **numData, long** catData, long noOfNumDim, long noOfCatDim)
{
	long i, j;

	long aNo;

	long scale = 10000;

	for (i = 0; i < noOfData; i++)
	{

		for (j = 0; j < noOfCatDim; j++)
		{
			fprintf(filePtr, "%ld ", catData[i][j]);
		}

		for(j = 0; j < noOfNumDim; j++ ){

				if (numData[i][j] == 1.0)
				{
					aNo = scale-1;
				}
				else
				{
					aNo = (long) (numData[i][j]*scale);
				}
				fprintf(filePtr, "%ld", aNo);
				if (j != noOfNumDim -1)
				{
					fprintf(filePtr, " ");
				}
			}

		fprintf(filePtr, "\n");

		}
}

void OutputPackageData(FILE *filePtr, long noOfData, double **numData, long noOfNumDim, long** packageCount)
{
	long i, j;

	long aNo;

	long scale = num_Scale ;

	for (i = 0; i < noOfData; i++)
	{

		long count = 0;

		for (j = 0; j < noOfNumDim; j++)
		{
			if (numData[i][j] == 1.0)
			{
				aNo = (scale-1);
			}
			else
			{
				aNo = (long) ((numData[i][j]*scale));
			}

			fprintf(filePtr, "%ld", aNo);
			if (j != noOfNumDim -1)
			{
				fprintf(filePtr, " ");
			}
		}

		fprintf(filePtr, "\n");
	}
}

double** GeneratePackageData(long count, long tCount,long dimension, long noTable, long* packageCount, long** rule,double*** testData){

	double** data = new double*[count];
	long* comb = new long[noTable];
	double discount1 = 1.0;
	double discount2 = 1.0;

	for(long i=0;i<count;i++){

		data[i] = new double[dimension];

		for(long j=0;j<dimension; j++){
			data[i][j] = 0;
		}

		for(long j=0;j < noTable; j++){
			comb[j] = random() % 100000;
		}
		//now we have a bunch of tables

		if(i % 200 == 0){
			discount2 = RandomNormal(1.0,variance);
			discount1 = RandomNormal(discount,variance);
		}

		for(long j=0;j<dimension;j++)
		{
			for(long k=0;k<packageCount[j];k++){
				long t = rule[j][2*k]-1;
				long d = rule[j][2*k+1];
				data[i][j] = data[i][j] + testData[t][comb[t]][d];
			}

			if(packageCount[j]>1){
				data[i][j] *= discount2;
			}
			else{
				data[i][j] *= discount1;
			}

		}

	}

	return data;
}

void GenerateData(long nMergT, long MergT, long MergP, long noOfTable, char Distribution, long tNo, long pNo, long** pCount, long*** rule)
//   ============
// generierte eine Datei mit zuf�lligen Daten
{
		double ***numData = new double**[noOfTable];
		double ***testData = new double**[noOfTable];

		double **package;

		switch (Distribution) {
		case 'E':
		case 'e': Distribution = 'E'; break;
		case 'C':
		case 'c': Distribution = 'C'; break;
		case 'A':
		case 'a': Distribution = 'A'; break;
		//default: printf("Ung�ltige Verteilung.\n"); return;
			default: printf("Invalid distribution.\n"); return;
		}

		FILE* f[noOfTable+1];

		for(long i=0;i<noOfTable+1;i++){
			char* fileName = new char[20];
			sprintf(fileName, "output/table%ld", i);
			f[i] = fopen(fileName, "w+");
			delete[] fileName;
		}


		if (f == NULL) {
			//printf("Kann Datei \"%s\" nicht anlegen.\n",FileName);
			printf("File \"%s\" cannot be opened.\n","data.txt");
			return;
		}

		for(long i=0;i<noOfTable;i++){

			switch (Distribution) {
			case 'E':
				testData[i] = GenerateDataEqually(100000,nMergT + MergT);
				break;
			case 'C':
				testData[i] = GenerateDataCorrelated(100000,nMergT + MergT);
				break;
			case 'A':
				testData[i] = GenerateDataAnticorrelated(100000,nMergT + MergT);
				break;
			}

		}

		long PackageDim = nMergT * noOfTable + MergP;

		package = GeneratePackageData(pNo,tNo,PackageDim,noOfTable, *pCount,*rule,testData);

		OutputPackageData(f[0], pNo, package, PackageDim, pCount);


		for(long i=0;i<noOfTable;i++){

			switch (Distribution) {
			case 'E':
				numData[i] = GenerateDataEqually(tNo,nMergT + MergT);
				break;
			case 'C':
				numData[i] = GenerateDataCorrelated(tNo,nMergT + MergT);
				break;
			case 'A':
				numData[i] = GenerateDataAnticorrelated(tNo,nMergT + MergT);
				break;
			}

		}


		for(long i=0;i<noOfTable;i++){
			OutputData(f[i+1], tNo, numData[i], nMergT + MergT);
		}

		for(long j=0;j>pNo;j++){
			delete[] package[j];
		}

		delete[] package;

		for(long j=0;j<noOfTable;j++){
			free(testData[j]);
			free(numData[j]);
		}

		delete[] numData;
		delete[] testData;


		for(long i=0;i<noOfTable+1;i++){
			fclose(f[i]);
		}

}

void genPoisson(double lambda){

	long k = 1;

	double result;

	if(poissonDistribution == NULL){

		poissonDistribution = new vector<double>;
		result = exp(-lambda);

		while(k<22){
			result = result * lambda / k;
			poissonDistribution->push_back(result);
			k++;
		}

	}

}

double ranPoisson(){

	if(poissonDistribution == NULL){
		genPoisson(1);
	}

	while(true){

	long i = rand() % 20 ;
	double threhold = RandomEqual(0,1);

	if(threhold < (*poissonDistribution)[i]){
		return i+1;
	}

	}

	return 0;
}

void GenRule(long nMergT, long MergT,long noTable,long MergP, long** pcArray, long*** rule){

	long packnum = noTable*nMergT + MergP;
	(*pcArray) = new long[packnum];
	memset(*pcArray,0,sizeof(long)*packnum);
	(*rule) = new long*[packnum];



	for(long k=0;k<packnum;k++){
		(*rule)[k] = new long[100];
	}

	for(long i=0;i<noTable;i++){

		for(long j=0;j<nMergT;j++){
			(*rule)[i*nMergT+j][0] = i + 1;
			(*rule)[i*nMergT+j][1] = j;
			(*pcArray)[i*nMergT+j] = 1;
		}

		for(long j=0;j<MergT;j++){
			long pick = rand() % MergP;
			long current = 2 * (*pcArray)[noTable*nMergT+pick];
			(*rule)[noTable*nMergT+pick][current] = i+1;
			(*rule)[noTable*nMergT+pick][current+1] = nMergT+j;
			(*pcArray)[noTable*nMergT+pick]++;
		}
	}

}

bool checkrule(long*** rule, long** pcArray, long n, long m){

for(long i=0;i<n;i++){
	if((*pcArray)[i] != 1){
		return false;
	}
}

for(long i=0;i<m;i++){
	if((*pcArray)[n+i] <= 1){
		return false;
	}

	for(long j=0;j<2*(*pcArray)[n+i];j+=2){
		for(long k=0;k<j;k+=2){
			if((*rule)[n+i][j]==(*rule)[n+i][k]){
				return false;
			}
		}
	}

}

return true;


}

void ruleToFile(long noOfTable, long nMergT, long MergT,long MergP,long tTuple, long pTuple, long** packageCount, long*** rule)
{
	FILE* fptr = fopen("output/config","w");

	if(fptr == NULL){
		printf("error in ruleToFile");
	}

	fprintf(fptr,"%d\n", noOfTable+1);
	fprintf(fptr,"%d %d %d\n",nMergT * noOfTable + MergP,pTuple,nMergT * noOfTable );

	for(long i=0;i<noOfTable;i++){
		fprintf(fptr,"%d %d %d\n",nMergT + MergT,tTuple,nMergT);
	}

	for(long i=0;i<noOfTable*nMergT+MergP;i++){
		fprintf(fptr,"%d ",(*packageCount)[i]);
	}

	fprintf(fptr,"\n");

	for(long i=0;i<noOfTable*nMergT+MergP;i++){
		for(long j=0;j<(*packageCount)[i]*2;j++){
			fprintf(fptr,"%d ",(*rule)[i][j]);
		}

		fprintf(fptr,"\n");
	}

	fclose(fptr);

}

long TE, TP, NA;

int main(int argc, char** argv){

	srand(time(0));

	if (argc != 4){
		printf("<|T_E| x k>, <|T_P| x k>, <no of attribute>");
	}
	else{
		TE = atoi(argv[1]) * 1000;
		TP = atoi(argv[2]) * 1000;
		NA = atoi(argv[3]);
	}

	int noCatDim = 0;//NA - 2;
	int noNumDim = NA - noCatDim;
	int noClass = 10;
	int Theta = 0.05;

	FILE* pFile = fopen("new_table.txt","w");
	fprintf(pFile,"%ld %ld\n", TP, NA);
	double** data = GenerateDataAnticorrelated(TP, noNumDim );
	long** catdata = GenerateCatData(TP, noCatDim, noClass, Theta);

	OutputData3(pFile, TP, data, catdata, noNumDim, noCatDim);

	fclose(pFile);

	pFile = fopen("exsiting_table.txt","w");
	fprintf(pFile,"%ld %ld\n", TE, NA);

	data = GenerateDataAnticorrelated(TE, noNumDim );
	catdata = GenerateCatData(TE, noCatDim, noClass, Theta);

	OutputData3(pFile, TE, data, catdata, noNumDim, noCatDim);
	fclose(pFile);

	return 0;
}

int main2(int argc, char** argv)
//  ====
// main program
{
	srand(time(0));

	if (argc != 11) {
		printf("generate - Generating data Set for Skyline\n\n");
		printf("Syntax: generate <No. Of tables> <No. of Num non-mergable Dims in table> <No. of mergable in table><No. of mergable Dims in package> <Distribution> <No. Of Tuples> <No. of Tuples in package> <No. of Merg in package> <Scale of Num> <mean of discount> <variance of discount> \n");
		printf("        \n\n");
	} else {
		long noOfTable = atoi(argv[1]);
		long nMergT = atoi(argv[2]);
		long MergT = atoi(argv[3]);
		long MergP = atoi(argv[4]);
 		char Distribution = argv[5][0];
		long tableTuple = atoi(argv[6]) * 1000;
		long packTuple = atoi(argv[7]) * 1000;
		long Scale = atoi(argv[8]);
		long dcnt = atoi(argv[9]);
		long var = atoi(argv[10]);

	long* pCount;
	long** rule;

	discount = dcnt / 100.0;

	variance = var / 100.0;

	if(MergT*noOfTable < MergP * 2){
		printf("MergP should be less than MergT*noOfTable/2");
		return 0;
	}

	if(MergT > MergP ){
		printf("MergP should be greater than MergT");
		return 0;
	}

	do{
		GenRule(nMergT,MergT,noOfTable,MergP, &pCount,&rule);
	}while(!checkrule(&rule, &pCount,nMergT*noOfTable,MergP));


	ruleToFile(noOfTable,nMergT,MergT,MergP,tableTuple,packTuple,&pCount,&rule);

	num_Scale = Scale;

	GenerateData(nMergT, MergT,MergP, noOfTable, Distribution, tableTuple, packTuple, &pCount, &rule);

	delete[] pCount;


	for(long k=0;k<noOfTable*nMergT+MergP;k++){
		delete[] rule[k];
	}

	delete[] rule;

	}
	return 0;
}

